/*
 * ------------------------------------------------------------------------
 *
 *  Copyright by KNIME AG, Zurich, Switzerland
 *  Website: http://www.knime.com; Email: contact@knime.com
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License, Version 3, as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, see <http://www.gnu.org/licenses>.
 *
 *  Additional permission under GNU GPL version 3 section 7:
 *
 *  KNIME interoperates with ECLIPSE solely via ECLIPSE's plug-in APIs.
 *  Hence, KNIME and ECLIPSE are both independent programs and are not
 *  derived from each other. Should, however, the interpretation of the
 *  GNU GPL Version 3 ("License") under any applicable laws result in
 *  KNIME and ECLIPSE being a combined program, KNIME AG herewith grants
 *  you the additional permission to use and propagate KNIME together with
 *  ECLIPSE with only the license terms in place for ECLIPSE applying to
 *  ECLIPSE and the GNU GPL Version 3 applying for KNIME, provided the
 *  license terms of ECLIPSE themselves allow for the respective use and
 *  propagation of ECLIPSE together with KNIME.
 *
 *  Additional permission relating to nodes for KNIME that extend the Node
 *  Extension (and in particular that are based on subclasses of NodeModel,
 *  NodeDialog, and NodeView) and that only interoperate with KNIME through
 *  standard APIs ("Nodes"):
 *  Nodes are deemed to be separate and independent programs and to not be
 *  covered works.  Notwithstanding anything to the contrary in the
 *  License, the License does not apply to Nodes, you are not required to
 *  license Nodes under the License, and you are granted a license to
 *  prepare and propagate Nodes, in each case even if such Nodes are
 *  propagated with or for interoperation with KNIME.  The owner of a Node
 *  may freely choose the license terms applicable to such Node, including
 *  when such Node is propagated with or for interoperation with KNIME.
 * ---------------------------------------------------------------------
 *
 * History
 *   Aug 9, 2018 (Julian Bunzel): created
 */
package org.knime.ext.textprocessing.language.turkish.nodes.preprocessing.zemberekstemmer;

import org.knime.core.data.DataTableSpec;
import org.knime.core.node.InvalidSettingsException;
import org.knime.core.node.NodeSettingsRO;
import org.knime.core.node.NodeSettingsWO;
import org.knime.core.node.defaultnodesettings.SettingsModelBoolean;
import org.knime.ext.textprocessing.language.turkish.util.ZemberekUtils;
import org.knime.ext.textprocessing.nodes.preprocessing.SentencePreprocessing;
import org.knime.ext.textprocessing.nodes.preprocessing.StreamableFunctionSentencePreprocessingNodeModel;

/**
 * The {@code NodeModel} for the Zemberek Stemmer node.
 *
 * @author Julian Bunzel, KNIME GmbH, Berlin, Germany
 */
final class ZemberekStemmerNodeModel extends StreamableFunctionSentencePreprocessingNodeModel {

    /** The default setting for maintaining the case of the stem. */
    private static final boolean DEF_MAINTAIN_CASE = true;

    /** The configuration key for the "maintain case" option. */
    private static final String CFG_KEY_MAINTAIN_CASE = "maintain_case";

    /**
     * Creates and returns a new instance of {@link SettingsModelBoolean} containing the boolean value for maintaining
     * the letter case of stems.
     *
     * @return Returns a new instance of {@code SettingsModelBoolean}
     */
    static final SettingsModelBoolean getMaintainCaseModel() {
        return new SettingsModelBoolean(CFG_KEY_MAINTAIN_CASE, DEF_MAINTAIN_CASE);
    }

    /** The {@link SettingsModelBoolean} for preservation of letter case. */
    private final SettingsModelBoolean m_maintainCase = getMaintainCaseModel();

    /**
     * {@inheritDoc}
     */
    @Override
    protected void internalConfigure(final DataTableSpec[] inSpecs) throws InvalidSettingsException {
        ZemberekUtils.verifyTokenizer(inSpecs[0], getDocumentColumn()).ifPresent(msg -> setWarningMessage(msg));
    }

    /**
     * {@inheritDoc}
     */
    @Override
    protected SentencePreprocessing createPreprocessing() throws Exception {
        return new WrappedZemberekStemmer(m_maintainCase.getBooleanValue());
    }

    /**
     * {@inheritDoc}
     */
    @Override
    protected void saveSettingsTo(final NodeSettingsWO settings) {
        super.saveSettingsTo(settings);
        m_maintainCase.saveSettingsTo(settings);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    protected void loadValidatedSettingsFrom(final NodeSettingsRO settings) throws InvalidSettingsException {
        super.loadValidatedSettingsFrom(settings);
        m_maintainCase.loadSettingsFrom(settings);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    protected void validateSettings(final NodeSettingsRO settings) throws InvalidSettingsException {
        super.validateSettings(settings);
        m_maintainCase.validateSettings(settings);
    }
}
