/*
 * ------------------------------------------------------------------------
 *
 *  Copyright by KNIME AG, Zurich, Switzerland
 *  Website: http://www.knime.com; Email: contact@knime.com
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License, Version 3, as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, see <http://www.gnu.org/licenses>.
 *
 *  Additional permission under GNU GPL version 3 section 7:
 *
 *  KNIME interoperates with ECLIPSE solely via ECLIPSE's plug-in APIs.
 *  Hence, KNIME and ECLIPSE are both independent programs and are not
 *  derived from each other. Should, however, the interpretation of the
 *  GNU GPL Version 3 ("License") under any applicable laws result in
 *  KNIME and ECLIPSE being a combined program, KNIME AG herewith grants
 *  you the additional permission to use and propagate KNIME together with
 *  ECLIPSE with only the license terms in place for ECLIPSE applying to
 *  ECLIPSE and the GNU GPL Version 3 applying for KNIME, provided the
 *  license terms of ECLIPSE themselves allow for the respective use and
 *  propagation of ECLIPSE together with KNIME.
 *
 *  Additional permission relating to nodes for KNIME that extend the Node
 *  Extension (and in particular that are based on subclasses of NodeModel,
 *  NodeDialog, and NodeView) and that only interoperate with KNIME through
 *  standard APIs ("Nodes"):
 *  Nodes are deemed to be separate and independent programs and to not be
 *  covered works.  Notwithstanding anything to the contrary in the
 *  License, the License does not apply to Nodes, you are not required to
 *  license Nodes under the License, and you are granted a license to
 *  prepare and propagate Nodes, in each case even if such Nodes are
 *  propagated with or for interoperation with KNIME.  The owner of a Node
 *  may freely choose the license terms applicable to such Node, including
 *  when such Node is propagated with or for interoperation with KNIME.
 * ---------------------------------------------------------------------
 *
 * History
 *   Feb 8, 2019 (julian): created
 */
package org.knime.ext.textprocessing.nodes.mining.relations;

import java.util.Arrays;
import java.util.Collections;
import java.util.List;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import org.knime.core.data.DataCell;
import org.knime.core.data.DataType;
import org.knime.core.data.MissingCell;
import org.knime.core.data.def.DoubleCell;
import org.knime.core.data.def.StringCell;

/**
 * Class to store results from relation extraction and convert them to DataCells.
 *
 * @author Julian Bunzel, KNIME GmbH, Berlin, Germany
 */
public class ExtractionResult {

    /**
     * List of {@link DataCell DataCells} from relation extraction.
     */
    private final List<DataCell> m_dataCells;

    /**
     * Creates a new instance of {@code ExtractionResult} with missing cells only.
     *
     * @param message Message shown as tool tip for missing values.
     */
    private ExtractionResult(final String message) {
        m_dataCells = Collections.unmodifiableList(
            Stream.generate(() -> new MissingCell(message)).limit(4).collect(Collectors.toList()));
    }

    /**
     * Creates a new instance of {@code ExtractionResult} containing the results from relation extraction.
     *
     * @param subject The subject.
     * @param relation The predicate.
     * @param object The object.
     * @param confidence The confidence.
     */
    public ExtractionResult(final String subject, final String relation, final String object, final Double confidence) {
        m_dataCells = Collections
            .unmodifiableList(Arrays.asList(subject != null ? new StringCell(subject) : DataType.getMissingCell(),
                relation != null ? new StringCell(relation) : DataType.getMissingCell(),
                object != null ? new StringCell(object) : DataType.getMissingCell(),
                confidence != null ? new DoubleCell(confidence) : DataType.getMissingCell()));
    }

    /**
     * Returns an {@code ExtractionResult} without any data.
     *
     * @param message Message shown as tool tip for missing values.
     * @return A new instance of {@code ExtractionResult} without any data.
     */
    public static final ExtractionResult getEmptyResult(final String message) {
        return new ExtractionResult(message);
    }

    /**
     * Returns the relation extraction results as unmodifiable list of {@link DataCell DataCells}.
     *
     * @return An unmodifiable list of {@code DataCells}.
     */
    public final List<DataCell> getDataCells() {
        return m_dataCells;
    }
}
